##
# Copyright 2016
# This file is part of openvim
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
##

import os
import json
import time
import subprocess
from git import Repo as gitrepo
from shutil import rmtree
from urllib.parse import urljoin

from charms.reactive import when, when_not, set_state
from charmhelpers.core.templating import render
from charmhelpers.core.hookenv import (
    application_version_set,
    config,
    leader_set,
    leader_get,
    status_set,
    unit_public_ip,
    open_port,
)
from charmhelpers.core.unitdata import kv
from charmhelpers.core.host import (
    symlink,
    mkdir,
    chownr,
    service_start,
)
from charmhelpers.contrib.unison import (
    create_private_key,
    create_public_key,
    ensure_user,
)
from charms.rest import RESTClient

cfg = config()

USER = "openvim"

# TODO: detect virtual machines and fail if they're not supported


def sh(cmd):
    return subprocess.check_output(cmd, shell=True)


def sh_as_openvim(cmd):
    return sh('sudo -iu openvim ' + cmd)


def create_openvim_user():
    status_set("maintenance", "Creating OpenVIM user")
    ensure_user('openvim')


def initialize_openvim_database(db):
    status_set("maintenance", "Initializing OpenVIM database")
    sh_as_openvim("/opt/openvim/database_utils/init_vim_db.sh -u %s -p %s -d %s -h %s" % (
        db.user(),
        db.password(),
        db.database(),
        db.host()
    ))


def generate_ssh_key():
    status_set("maintenance", "Generating ssh key")
    user = "openvim"
    folder = "/home/%s/.ssh" % user
    mkdir(folder, owner=user, group=user, perms=0o700)
    private_path = "%s/id_rsa" % folder
    public_path = "%s.pub" % private_path
    create_private_key(user, private_path)
    create_public_key(user, private_path, public_path)


def add_openvim_to_path():
    status_set("maintenance", "Adding OpenVIM to path")
    symlink(
        '/opt/openvim/scripts/service-vim.sh',
        '/usr/bin/service-openvim')
    symlink('/opt/openvim/openvim', '/usr/bin/openvim')


def download_openvim():
    status_set("maintenance", "Downloading OpenVIM")
    if os.path.isdir("/opt/openvim"):
        rmtree("/opt/openvim")
    repo = gitrepo.clone_from(
        cfg['repository'],
        '/opt/openvim',
        branch=cfg['branch'],
    )
    chownr(
        '/opt/openvim',
        owner='openvim',
        group='openvim',
        follow_links=False,
        chowntopdir=True
    )

    # Set the application version, based the sha1sum of HEAD
    revision = repo.rev_parse("HEAD").name_rev[:7]
    application_version_set(revision)


def configure_openvim(db):
    status_set("maintenance", "Configuring OpenVIM")
    render(
        source="openvimd.cfg",
        target="/opt/openvim/openvimd.cfg",
        owner="openvim",
        perms=0o664,
        context={"db": db, "mode": cfg['mode']}
    )


# TODO: possibly combine all of these create functions?
def create_tenant():
    status_set("maintenance", "Creating tenant")
    render(source="tenant.yaml", target="/tmp/tenant.yaml", owner="openvim", perms=0o664, context={})
    cmd = 'openvim tenant-create /tmp/tenant.yaml'
    tenant_uuid = sh_as_openvim(cmd).split()[0]
    tenant_uuid = str(tenant_uuid, 'utf-8')
    leader_set({'tenant': tenant_uuid})
    return tenant_uuid


def create_image():
    status_set("maintenance", "Creating image")
    render(source="image.yaml", target="/tmp/image.yaml", owner="openvim", perms=0o664, context={})
    cmd = 'openvim image-create /tmp/image.yaml'
    image_uuid = sh_as_openvim(cmd).split()[0]
    image_uuid = str(image_uuid, 'utf-8')
    return image_uuid


def create_flavor():
    status_set("maintenance", "Creating flavor")
    render(source="flavor.yaml", target="/tmp/flavor.yaml", owner="openvim", perms=0o664, context={})
    cmd = 'openvim flavor-create /tmp/flavor.yaml'
    flavor_uuid = sh_as_openvim(cmd).split()[0]
    flavor_uuid = str(flavor_uuid, 'utf-8')
    return flavor_uuid


# TODO: especially combine these stupid network functions
def create_default_network():
    status_set("maintenance", "Creating default network")
    render(source="net-default.yaml", target="/tmp/net-default.yaml", owner="openvim", perms=0o664, context={})
    cmd = 'openvim net-create /tmp/net-default.yaml'
    net_default_uuid = sh_as_openvim(cmd).split()[0]
    net_default_uuid = str(net_default_uuid, 'utf-8')
    return net_default_uuid


def create_virbr_network():
    status_set("maintenance", "Creating virbr0 network")
    render(source="net-virbr0.yaml", target="/tmp/net-virbr0.yaml", owner="openvim", perms=0o664, context={})
    cmd = 'openvim net-create /tmp/net-virbr0.yaml'
    net_virbr0_uuid = sh_as_openvim(cmd).split()[0]
    net_virbr0_uuid = str(net_virbr0_uuid, 'utf-8')
    return net_virbr0_uuid


def create_vm_yaml(image_uuid, flavor_uuid, net_default_uuid, net_virbr0_uuid):
    status_set("maintenance", "Creating default VM yaml file")
    render(
        source="server.yaml",
        target="/tmp/server.yaml",
        owner="openvim",
        perms=0o664,
        context={
            "image_uuid": image_uuid,
            "flavor_uuid": flavor_uuid,
            "net_default_uuid": net_default_uuid,
            "net_virbr0_uuid": net_virbr0_uuid
        }
    )


def create_sane_defaults():
    tenant_uuid = create_tenant()
    add_openvim_tenant_env_var(tenant_uuid)
    image_uuid = create_image()
    flavor_uuid = create_flavor()
    net_default_uuid = create_default_network()
    net_virbr0_uuid = create_virbr_network()
    create_vm_yaml(
        image_uuid=image_uuid,
        flavor_uuid=flavor_uuid,
        net_default_uuid=net_default_uuid,
        net_virbr0_uuid=net_virbr0_uuid
    )


def install_openvim_service():
    status_set("maintenance", "Installing OpenVIM service")
    if not os.path.exists('/etc/systemd/system'):
        os.makedirs('/etc/systemd/system')

    # TODO: If you render a template that already exists,
    # we need to call `systemctl daemon-reload`
    render(
        source="openvim.service",
        target="/etc/systemd/system/openvim.service",
        owner="root",
        perms=0o644,
        context={}
    )


def add_openvim_tenant_env_var(tenant_uuid):
    status_set("maintenance", "Adding OPENVIM_TENANT environment variable")
    env_line = 'export OPENVIM_TENANT=%s\n' % tenant_uuid
    with open('/home/openvim/.profile', 'w+') as f:
        lines = f.readlines()
        for line in lines:
            if env_line == line:
                return
        f.seek(0)
        f.truncate()
        for line in lines:
            f.write(line)
        f.write(env_line)


def openvim_running():
    try:
        sh_as_openvim('openvim tenant-list')
        return True
    except:
        return False


def start_openvim():
    status_set("maintenance", "Starting OpenVIM")
    service_start('openvim')
    t0 = time.time()
    while not openvim_running():
        if time.time() - t0 > 60:
            raise Exception('Failed to start openvim.')
        time.sleep(0.25)


@when_not('db.available')
def not_ready():
    status_set('waiting', 'MySQL database required')


@when('db.available')
@when_not('openvim-controller.installed')
def install_openvim_controller(mysql):
    create_openvim_user()
    download_openvim()
    add_openvim_to_path()
    configure_openvim(mysql)
    initialize_openvim_database(mysql)
    generate_ssh_key()
    install_openvim_service()
    start_openvim()
    create_sane_defaults()
    status_set(
        'active',
        'Up on {host}:{port}'.format(
            host=unit_public_ip(),
            port='9080'))

    # TODO: Make API and Admin port configurable
    open_port(9080)
    open_port(9085)
    set_state('openvim-controller.installed')


@when('compute.connected', 'openvim-controller.installed')
def send_ssh_key(compute):
    with open('/home/openvim/.ssh/id_rsa.pub', 'r') as f:
        key = f.read().strip()
    compute.send_ssh_key(key)

@when('openvim-controller.installed)')
@when_not('compute.available')
def host_remove(compute):
    """Remove the compute nodes from the openvim controller"""
    cache = kv()

    # Connect to OpenVim's administrative port
    openvim = OpenVimApi("localhost", 9085)

    for node in compute.authorized_nodes():
        id = cache.get("compute:" + node['address'])
        if id:
            openvim.host_delete(id)
    status_set('waiting', 'MySQL database required')



@when('compute.available', 'openvim-controller.installed')
def host_add(compute):
    cache = kv()

    # This requires the administrative interface
    openvim = OpenVimApi("localhost", 9085)

    for node in compute.authorized_nodes():

        if cache.get("compute:" + node['address']):
            continue

        # We need to set a unique name, because IP isn't easily parsed from
        # the /hosts api
        name = node['address'].replace(".", "-")

        if name not in openvim.get_host_names():
            cmd = "ssh -n -o 'StrictHostKeyChecking no' %s@%s"
            sh_as_openvim(cmd % (node['user'], node['address']))

            # with open('/tmp/compute-0.yaml', 'w') as f:
            #     f.write(node['config'])

            resp = openvim.host_add(node['config'])
            if resp:
                id = resp['host']['id']
                cache.set('compute:' + node['address'], id)


@when('openvim-controller.available')
def openvim_available(openvim):
    openvim.configure(port=9080, user=leader_get('tenant'))


# TODO: move this to it's own layer, perhaps?
class OpenVimApi(RESTClient):
    """ A wrapper around the OpenVIM API"""
    uri = None

    def __init__(self, host, port=9080, ssl=False):
        if host and port:
            base = "http"
            if ssl:
                base = "https"

            self.uri = "{}://{}:{}".format(base, host, port)
        pass

    # Server primitives

    ###########
    # Tenants #
    ###########
    def get_tenants():
        pass

    def get_tenant(tenant_id):
        pass

    def set_tenants():
        pass

    def set_tenant(tenant_id, tenant):
        pass

    ###########
    # Flavors #
    ###########
    def get_flavors(tenant_id):
        pass

    def get_flavor(tenant_id, flavor_id):
        pass

    def set_flavors(tenant_id, flavors):
        pass

    def set_flavor(tenant_id, flavor_id, flavor):
        pass

    def delete_flavor(tenant_id, flavor_id):
        pass

    ##########
    # Images #
    ##########
    def get_images(tenant_id):
        pass

    def get_image(tenant_id, image_id):
        pass

    def set_image(tenant_id, image_id, image):
        pass

    def delte_image(tenant_id, image_id):
        pass

    ###########
    # Servers #
    ###########
    def get_servers(tenant_id):
        pass

    def get_server(tenant_id, server_id):
        pass

    def set_server(tenant_id, server_id, server):
        pass

    def set_server_action(tenant_id, server_id, action):
        pass

    def delete_server(tenant_id, server_id):
        pass

    # Network primitives


    # Administrative primitives

    def hosts(self):
        """Get a list of all available hosts"""
        if self.uri:
            url = urljoin(self.uri, "/openvim/hosts")
            r = self.get(url)
            if r.status_code == 200:
                return r.json()
            else:
                return {}

    def get_host_names(self):
        names = []
        hosts = self.hosts()
        for host in hosts['hosts']:
            names.append(host['name'])
        return names

    def get_host_ids(self):
        ids = []
        hosts = self.hosts()
        for host in hosts['hosts']:
            ids.append(host['id'])
        return ids

    def host_add(self, data):
        """Add a host to OpenVim and return its id"""
        if self.uri and data:
            url = urljoin(self.uri, "/openvim/hosts")
            headers = {'Content-Type': 'application/yaml'}
            r = self.post(url, data, headers=headers)
            if r.status_code == 200:
                return r.json()
            else:
                return None

    def host_delete(self, id):
        """Delete a host from OpenVim"""
        if self.uri and id:
            url = urljoin(self.uri, "/openvim/hosts/{}".format(id))
            self.delete(url)


    # OpenFlow rules

    # External ports
